@@ -1,27 +1,46 @@
 /* SASL server API implementation
  * Tim Martin
  */
-/***********************************************************
-        Copyright 1998 by Carnegie Mellon University
+/* 
+ * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ */
 
-                      All Rights Reserved
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for any purpose and without fee is hereby granted,
-provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in
-supporting documentation, and that the name of CMU not be
-used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.
-
-CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
-ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
-CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
-ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
-ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
-SOFTWARE.
-******************************************************************/
 
 /* checkpw stuff */
 
@@ -52,9 +71,9 @@ SOFTWARE.
 #include <sys/types.h>
 #include <ctype.h>
 
-#ifndef SASL_MINIMAL_SERVER
+#ifdef HAVE_PWD_H
 #include <pwd.h>
-#endif /* SASL_MINIMAL_SERVER */
+#endif /* HAVE_PWD_H */
 #ifdef HAVE_CRYPT_H
 #ifndef HAVE_KRB
 #include <crypt.h>
@@ -207,15 +226,19 @@ static int use_key(char *user __attribute__((unused)),
  * 0 for failure.  On failure, 'reply' is filled in with a pointer to
  * the reason.
  */
-int _sasl_kerberos_verify_password(sasl_conn_t *conn,
-				   const char *user, const char *passwd,
-				   const char *service, const char **reply)
+static int kerberos_verify_password(sasl_conn_t *conn,
+				    const char *user, 
+				    const char *passwd,
+				    const char *service, 
+				    const char *user_realm 
+				               __attribute__((unused)), 
+				    const char **reply)
 {
     int result;
     des_cblock key;
     char tfname[40];
-    char realm[REALM_SZ];
-    char cell[REALM_SZ];
+    char realm[REALM_SZ+1];
+    char cell[REALM_SZ+1];
     char hostname[MAXHOSTNAMELEN+1];
     char phost[MAXHOSTNAMELEN+1];
     char oldtkfile[MAXPATHLEN];
@@ -243,8 +266,8 @@ int _sasl_kerberos_verify_password(sasl_conn_t *conn,
     if (krb_get_lrealm(realm, 1)) return SASL_FAIL;
 
     /* after this point, we need to "goto fini" on failure */
-    strcpy(oldtkfile, tkt_string());
-    sprintf(tfname, "/tmp/tkt_%d", getpid());
+    strcpy(oldtkfile, tkt_string()); /* krb gaurantees only MAXPATHLEN-1 length */
+    snprintf(tfname,sizeof(tfname), "/tmp/tkt_%d", getpid());
     krb_set_tkt_string(tfname);
 
     /* First try Kerberos string-to-key */
@@ -255,6 +278,7 @@ int _sasl_kerberos_verify_password(sasl_conn_t *conn,
 
     if (result == INTK_BADPW) {
 	/* Now try andrew string-to-key */
+	realm[REALM_SZ] = '\0';
 	strcpy(cell, realm);
 	lcase(cell);
 	krb_afs_string_to_key(passwd, key, cell);
@@ -273,7 +297,7 @@ int _sasl_kerberos_verify_password(sasl_conn_t *conn,
 
     /* Check validity of returned ticket */
     gethostname(hostname, sizeof(hostname));
-    strcpy(phost, krb_get_phost(hostname));
+    strncpy(phost, krb_get_phost(hostname), sizeof(phost)-1);
 
     result = krb_mk_req(&authent, (char *) service, phost, realm, 0);
 
@@ -285,7 +309,8 @@ int _sasl_kerberos_verify_password(sasl_conn_t *conn,
     }
     strcpy(instance, "*");
 
-    result = krb_rd_req(&authent, (char *) service, instance, 0L, &kdata, srvtab); 
+    result = krb_rd_req(&authent, (char *) service, instance, 0L, &kdata, 
+			srvtab); 
 
     memset(&authent, 0, sizeof(authent));
     memset(kdata.session, 0, sizeof(kdata.session));
@@ -311,15 +336,15 @@ int _sasl_kerberos_verify_password(sasl_conn_t *conn,
 
 #endif /* HAVE_KRB */
 
-#ifndef SASL_MINIMAL_SERVER
-
-int _sasl_shadow_verify_password(sasl_conn_t *conn __attribute__((unused)),
-				 const char *userid,
-				 const char *password,
-				 const char **reply __attribute__((unused)) )
-{
 #ifdef HAVE_GETSPNAM
-
+static int shadow_verify_password(sasl_conn_t *conn __attribute__((unused)),
+				  const char *userid,
+				  const char *password,
+				  const char *service __attribute__((unused)),
+				  const char *user_realm 
+				               __attribute__((unused)), 
+				  const char **reply)
+{
   char *salt;
   char *crypted;
   struct spwd *spwd;
@@ -338,17 +363,17 @@ int _sasl_shadow_verify_password(sasl_conn_t *conn __attribute__((unused)),
     return SASL_BADAUTH;	/* we lose. */
 
   return SASL_OK;
-
-#else  /* HAVE_GETSPNAM */
-  return SASL_FAIL;
-#endif
-
 }
+#endif /* HAVE_GETSPNAM */
 
-int _sasl_passwd_verify_password(sasl_conn_t *conn __attribute__((unused)),
-				 const char *userid,
-				 const char *password,
-				 const char **reply)
+#ifdef HAVE_GETPWNAM
+static int passwd_verify_password(sasl_conn_t *conn __attribute__((unused)),
+				  const char *userid,
+				  const char *password,
+				  const char *service __attribute__((unused)),
+				  const char *user_realm 
+				               __attribute__((unused)), 
+				  const char **reply)
 {
   struct passwd *pwd;
   char *salt;
@@ -371,7 +396,8 @@ int _sasl_passwd_verify_password(sasl_conn_t *conn __attribute__((unused)),
 
   return SASL_BADAUTH;
 }
-#endif /* SASL_MINIMAL_SERVER */
+#endif /* HAVE_GETPWNAM */
+
 #ifdef HAVE_PAM
 struct sasl_pam_data {
     const char *userid;
@@ -434,10 +460,12 @@ static struct pam_conv my_conv = {
     NULL			/* appdata_ptr */
 };
 
-int _sasl_PAM_verify_password(sasl_conn_t *conn __attribute__((unused)),
-			      const char *userid, const char *password,
-			      const char *service,
-			      const char **reply)
+static int pam_verify_password(sasl_conn_t *conn __attribute__((unused)),
+			       const char *userid, 
+			       const char *password,
+			       const char *service,
+			       const char *user_realm __attribute__((unused)), 
+			       const char **reply)
 {
     pam_handle_t *pamh;
     struct sasl_pam_data pd;
@@ -526,13 +554,9 @@ static int parseuser(char **user, char **realm, const char *user_realm,
 	/* otherwise, we gotta get it from the user */
 	r = strchr(input, '@');
 	if (!r) {
-	    *realm = sasl_ALLOC(1);
-	    if (*realm) {
-		(*realm)[0] = '\0';
-		ret = SASL_OK;
-	    } else {
-		ret = SASL_NOMEM;
-	    }
+	    /* hmmm, the user didn't specify a realm */
+	    /* we'll default to the serverFQDN */
+	    ret = _sasl_strdup(serverFQDN, realm, NULL);
 	    if (ret == SASL_OK) {
 		ret = _sasl_strdup(input, user, NULL);
 	    }
@@ -542,7 +566,7 @@ static int parseuser(char **user, char **realm, const char *user_realm,
 	    *--r = '\0';
 	    *user = sasl_ALLOC(r - input + 1);
 	    if (*user) {
-		strcpy(*user, input);
+		strncpy(*user, input, r - input +1);
 	    } else {
 		ret = SASL_NOMEM;
 	    }
@@ -553,58 +577,62 @@ static int parseuser(char **user, char **realm, const char *user_realm,
     return ret;
 }
 
-int _sasl_sasldb_verify_password(sasl_conn_t *conn,
-				 const char *userstr, const char *passwd,
-				 const char *user_realm, const char **reply)
+static int sasldb_verify_password(sasl_conn_t *conn,
+				  const char *userstr,
+				  const char *passwd,
+				  const char *service __attribute__((unused)),
+				  const char *user_realm, 
+				  const char **reply)
 {
     sasl_server_getsecret_t *getsec;
-    void *context;
-    int ret;
+    void *context = NULL;
+    int ret = SASL_FAIL;
     sasl_secret_t *secret = NULL;
     sasl_secret_t *construct = NULL;
-    char *userid, *realm;
-
-    VL(("Verifying password via sasldb\n"));
+    char *userid = NULL;
+    char *realm = NULL;
 
+    if (reply) { *reply = NULL; }
     if (!userstr || !passwd) {
 	return SASL_BADPARAM;
     }
-    if (reply) { *reply = NULL; }
     ret = parseuser(&userid, &realm, user_realm, conn->serverFQDN, userstr);
     if (ret != SASL_OK) {
-	VL(("Error parsing user\n"));
-	return ret;
+	/* error parsing user */
+	goto done;
     }
     ret = _sasl_getcallback(conn, SASL_CB_SERVER_GETSECRET, &getsec, &context);
     if (ret != SASL_OK) {
-	VL(("Error getting getsecret callback\n"));
-	return ret;
+	/* error getting getsecret callback */
+	goto done;
     }
 
     ret = getsec(context, "PLAIN", userid, realm, &secret);
     if (ret != SASL_OK) {
-	VL(("Error getting PLAIN secret for user %s in realm %s from db\n",userid, realm));
-	return ret;
+	/* error getting PLAIN secret */
+	goto done;
     }
 
     ret = _sasl_make_plain_secret(secret->data, passwd, strlen(passwd),
 				  &construct);
     if (ret != SASL_OK) {
-	sasl_free_secret(&secret);
-	if (construct) { sasl_free_secret(&construct); }
-	return ret;
+	goto done;
     }
 
     if (!memcmp(secret->data, construct->data, secret->len)) {
-	VL(("Password verified!\n"));
+	/* password verified! */
 	ret = SASL_OK;
     } else {
-	VL(("Passwords do not match"));
+	/* passwords do not match */
 	ret = SASL_BADAUTH;
     }
 
-    sasl_free_secret(&secret);
-    sasl_free_secret(&construct);
+ done:
+    if (userid) sasl_FREE(userid);
+    if (realm)  sasl_FREE(realm);
+
+    if (secret) sasl_free_secret(&secret);
+    if (construct) sasl_free_secret(&construct);
     return ret;
 }
 
@@ -617,7 +645,8 @@ int _sasl_sasldb_set_pass(sasl_conn_t *conn,
 			  int flags,
 			  const char **errstr)
 {
-    char *userid, *realm;
+    char *userid = NULL;
+    char *realm = NULL;
     int ret = SASL_OK;
 
     if (errstr) {
@@ -681,7 +710,8 @@ int _sasl_sasldb_set_pass(sasl_conn_t *conn,
 	if (sec) {
 	    sasl_free_secret(&sec);
 	}
-    } else {
+    } else { 
+	/* SASL_SET_DISABLE specified */
 	sasl_server_putsecret_t *putsec;
 	void *context;
 
@@ -690,9 +720,14 @@ int _sasl_sasldb_set_pass(sasl_conn_t *conn,
 	if (ret == SASL_OK) {
 	    ret = putsec(context, "PLAIN", userid, realm, NULL);
 	}
+	if (ret != SASL_OK) {
+	    _sasl_log(conn, SASL_LOG_ERR, NULL, ret, 0,
+		      "failed to disable account for %s: %z", userid);
+	}
     }
 
-    sasl_FREE(realm);
+    if (userid)   sasl_FREE(userid);
+    if (realm)    sasl_FREE(realm);
     return ret;
 }
 
@@ -756,9 +791,13 @@ static int retry_writev(int fd, struct iovec *iov, int iovcnt)
 
 
 /* pwcheck daemon-authenticated login */
-int _sasl_pwcheck_verify_password(sasl_conn_t *conn,
-				  const char *userid, const char *passwd,
-				  const char **reply)
+static int pwcheck_verify_password(sasl_conn_t *conn,
+				   const char *userid, 
+				   const char *passwd,
+				   const char *service __attribute__((unused)),
+				   const char *user_realm 
+				               __attribute__((unused)), 
+				   const char **reply)
 {
     int s;
     struct sockaddr_un srvaddr;
@@ -772,6 +811,8 @@ int _sasl_pwcheck_verify_password(sasl_conn_t *conn,
 
     if (reply) { *reply = NULL; }
 
+    if (strlen(PWCHECKDIR)+8+1 > sizeof(pwpath)) return SASL_FAIL;
+
     strcpy(pwpath, PWCHECKDIR);
     strcat(pwpath, "/pwcheck");
 
@@ -780,7 +821,7 @@ int _sasl_pwcheck_verify_password(sasl_conn_t *conn,
 
     memset((char *)&srvaddr, 0, sizeof(srvaddr));
     srvaddr.sun_family = AF_UNIX;
-    strcpy(srvaddr.sun_path, pwpath);
+    strncpy(srvaddr.sun_path, pwpath, sizeof(srvaddr.sun_path));
     r = connect(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr));
     if (r == -1) {
 	if (reply) { *reply = "cannot connect to pwcheck server"; }
@@ -813,3 +854,23 @@ int _sasl_pwcheck_verify_password(sasl_conn_t *conn,
 }
 
 #endif
+
+struct sasl_verify_password_s _sasl_verify_password[] = {
+    { "sasldb", &sasldb_verify_password },
+#ifdef HAVE_KRB
+    { "kerberos_v4", &kerberos_verify_password },
+#endif
+#ifdef HAVE_GETSPNAM
+    { "shadow", &shadow_verify_password },
+#endif
+#ifdef HAVE_GETPWNAM
+    { "passwd", &passwd_verify_password },
+#endif
+#ifdef HAVE_PAM
+    { "pam", &pam_verify_password },
+#endif
+#ifdef HAVE_PWCHECK
+    { "pwcheck", &pwcheck_verify_password },
+#endif
+    { NULL, NULL }
+};
