@@ -1,28 +1,47 @@
 /* CRAM-MD5 SASL plugin
  * Tim Martin 
- * $Id: cram.c,v 1.47 2000/02/23 02:50:01 leg Exp $
+ * $Id: cram.c,v 1.53 2000/03/12 02:20:00 leg Exp $
  */
-/***********************************************************
-        Copyright 1998 by Carnegie Mellon University
 
-                      All Rights Reserved
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for any purpose and without fee is hereby granted,
-provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in
-supporting documentation, and that the name of CMU not be
-used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.
-
-CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
-ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
-CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
-ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
-ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
-SOFTWARE.
-******************************************************************/
+/* 
+ * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -89,6 +108,7 @@ static int start(void *glob_context __attribute__((unused)),
   /* not used on server side */
   text->authid=NULL;
   text->password=NULL;
+  text->msgid = NULL;
   
   *conn=text;
 
@@ -188,7 +208,7 @@ static char * randomdigits(sasl_server_params_t *sparams)
       (temp[2] * 256) +
       (temp[3] );
 
-  ret = sparams->utils->malloc(15);
+  ret = sparams->utils->malloc(15); /* there's no way an unsigned can be longer than this right? */
   if (ret == NULL) return NULL;
   sprintf(ret, "%u", num);
 
@@ -222,13 +242,9 @@ static int parseuser(sasl_utils_t *utils,
 	/* otherwise, we gotta get it from the user */
 	r = strchr(input, '@');
 	if (!r) {
-	    *realm = utils->malloc(1);
-	    if (*realm) {
-		*realm[0] = '\0';
-		ret = SASL_OK;
-	    } else {
-		ret = SASL_NOMEM;
-	    }
+	    /* hmmm, the user didn't specify a realm */
+	    /* we'll default to the serverFQDN */
+	    ret = cram_strdup(utils, serverFQDN, realm, NULL);
 	    if (ret == SASL_OK) {
 		ret = cram_strdup(utils, input, user, NULL);
 	    }
@@ -343,9 +359,8 @@ static int server_continue_step (void *conn_context,
   context_t *text;
   text=conn_context;
 
-  if (errstr) {
-      *errstr = NULL;
-  }
+  if (errstr) *errstr = NULL;
+  if (clientinlen < 0) return SASL_BADPARAM;
 
   if (text->state==1)
   {    
@@ -370,7 +385,7 @@ static int server_continue_step (void *conn_context,
     if ((time==NULL) || (randdigits==NULL)) return SASL_NOMEM;
 
     /* allocate some space for the nonce */
-    *serverout=sparams->utils->malloc(200);
+    *serverout=sparams->utils->malloc(200+1);
     if (*serverout==NULL) return SASL_NOMEM;
 
     /* create the nonce */
@@ -404,12 +419,12 @@ static int server_continue_step (void *conn_context,
     char *authstr = NULL;
     sasl_secret_t *sec=NULL;
     int lup,pos;
-    int result;
+    int result = SASL_FAIL;
     sasl_server_getsecret_t *getsecret;
     void *getsecret_context;
 
     HMAC_MD5_CTX tmphmac;
-    char *digest_str;
+    char *digest_str = NULL;
     UINT4 digest[4];
 
     VL(("CRAM-MD5 Step 2\n"));
@@ -420,7 +435,7 @@ static int server_continue_step (void *conn_context,
     while ((pos>0) && (clientin[pos]!=' ')) {
 	pos--;
     }
-    if (pos==0) {
+    if (pos<=0) {
 	if (errstr) *errstr = "need authentication name";
 	return SASL_BADPROT;
     }
@@ -437,7 +452,7 @@ static int server_continue_step (void *conn_context,
 	      sparams->serverFQDN, authstr);
     sparams->utils->free(authstr);
     if (result != SASL_OK) {
-	return result;
+	goto done;
     }
 
     /* get callback so we can request the secret */
@@ -447,11 +462,12 @@ static int server_continue_step (void *conn_context,
 					 &getsecret_context);
     if (result != SASL_OK) {
 	VL(("result = %i trying to get secret callback\n",result));
-	return result;
+	goto done;
     }
     if (! getsecret) {
 	VL(("Received NULL getsecret callback\n"));
-	return SASL_FAIL;
+	result = SASL_FAIL;
+	goto done;
     }
 
     /* We use the user's CRAM secret which is kinda 1/2 way thru the
@@ -460,20 +476,17 @@ static int server_continue_step (void *conn_context,
     result = getsecret(getsecret_context, "CRAM-MD5", userid, realm, &sec);
     if (result == SASL_NOUSER || !sec) {
       if (errstr) *errstr = "no secret in database";
-      return SASL_NOUSER;
+      result = SASL_NOUSER;
+      goto done;
     }
     if (result != SASL_OK) {
-	sparams->utils->free(userid);
-	sparams->utils->free(realm);
-	return result;
+	goto done;
     }
 
     if (sec->len != sizeof(HMAC_MD5_STATE)) {
-      free_secret(sparams->utils, &sec);
-      free_string(sparams->utils, &userid);
-      free_string(sparams->utils, &realm);
       if (errstr) *errstr = "secret database corruption";
-      return SASL_FAIL;
+      result = SASL_FAIL;
+      goto done;
     }
 
     /* ok this is annoying:
@@ -492,22 +505,16 @@ static int server_continue_step (void *conn_context,
     /* convert to base 16 with lower case letters */
     digest_str = convert16((unsigned char *) digest, 4, sparams->utils);
 
-    /* free sec */
-    free_secret(sparams->utils,&sec);
-
     /* if same then verified 
      *  - we know digest_str is null terminated but clientin might not be
      */
     if (strncmp(digest_str,clientin+pos+1,strlen(digest_str))!=0) {
-	sparams->utils->free(userid);
 	if (errstr) {
 	    *errstr = "incorrect digest response";
-	}
-	free(digest_str);
-	return SASL_BADAUTH;
+	}	
+	result = SASL_BADAUTH;
+	goto done;
     }
-    free(digest_str);
-
     VL(("Succeeded!\n"));
 
     /* nothing more to do; authenticated 
@@ -516,10 +523,14 @@ static int server_continue_step (void *conn_context,
     oparams->doneflag=1;
 
     oparams->user = userid; /* set username */
+    userid = NULL; /* set to null so we don't free */
     oparams->realm = realm;
+    realm = NULL; /* set to null so we don't free */
     
-    result = cram_strdup(sparams->utils, userid, &(oparams->authid), NULL);
-    if (result!=SASL_OK) return result;
+    result = cram_strdup(sparams->utils, oparams->user, &(oparams->authid), NULL);
+    if (result!=SASL_OK) {
+	goto done;
+    }
 
     oparams->mech_ssf = 0;
     oparams->maxoutbuf = 0;
@@ -531,9 +542,17 @@ static int server_continue_step (void *conn_context,
     *serverout = NULL;
     *serveroutlen = 0;
 
+    result = SASL_OK;
+  done:
+
+    if (userid) sparams->utils->free(userid);
+    if (realm)  sparams->utils->free(realm);
+    if (sec) free_secret(sparams->utils, &sec);
+    if (digest_str)  sparams->utils->free(digest_str);
+
     text->state = 3; /* if called again will fail */
 
-    return SASL_OK;
+    return result;
   }
 
 
@@ -553,7 +572,7 @@ static int mechanism_db_filled(char *mech_name, sasl_utils_t *utils)
   int result;
   sasl_server_getsecret_t *getsecret;
   void *getsecret_context;
-  int version = -1;
+  long version = -1;
 
   /* get callback so we can request the secret */
   result = utils->getcallback(utils->conn,
@@ -591,7 +610,7 @@ static int mechanism_db_filled(char *mech_name, sasl_utils_t *utils)
 		 mech_name,
 		 SASL_FAIL,
 		 0,
-		 "CRAM-MD5 secrets database has incompatible version (%d). My version (%d)",
+		 "CRAM-MD5 secrets database has incompatible version (%ld). My version (%d)",
 		 version, CRAM_MD5_VERSION);
 
       return SASL_FAIL;
@@ -608,14 +627,13 @@ static int mechanism_db_filled(char *mech_name, sasl_utils_t *utils)
  * Note: this function is duplicated in multiple plugins. If you fix
  * something here please update the other files
  */
-
 static int mechanism_fill_db(char *mech_name, sasl_server_params_t *sparams)
 {
   int result;
   sasl_server_putsecret_t *putsecret;
   void *putsecret_context;
   sasl_secret_t *sec = NULL;
-  int version;
+  long version;
 
   /* don't do this again if it's already set */
   if (mydb_initialized == 1)
@@ -674,8 +692,8 @@ setpass(void *glob_context __attribute__((unused)),
     int result;
     sasl_server_putsecret_t *putsecret;
     void *putsecret_context;
-    char *user;
-    char *realm;
+    char *user = NULL;
+    char *realm = NULL;
 
     /* These need to be zero'ed out at the end */
     HMAC_MD5_STATE *md5state = NULL;
@@ -753,6 +771,8 @@ setpass(void *glob_context __attribute__((unused)),
 	memset(md5state, 0, sizeof(md5state));
 	sparams->utils->free(md5state);
     }
+    if (user) 	sparams->utils->free(user);
+    if (realm) 	sparams->utils->free(realm);
     return result;
 }
 
@@ -855,10 +875,10 @@ static int get_authid(sasl_client_params_t *params,
   if (prompt!=NULL)
   {
     /* copy it */
-    *authid=params->utils->malloc(strlen(prompt->result)+1);
+    *authid=params->utils->malloc(prompt->len+1);
     if ((*authid)==NULL) return SASL_NOMEM;
 
-    strcpy(*authid, prompt->result);
+    strncpy(*authid, prompt->result, prompt->len+1);
     return SASL_OK;
   }
 
@@ -1047,8 +1067,8 @@ static int c_continue_step (void *conn_context,
   oparams->maxoutbuf=0; /* no clue what this should be */
   oparams->encode=NULL;
   oparams->decode=NULL;
-  oparams->user="anonymous"; /* set username */
-  oparams->authid="anonymous";
+  oparams->user=NULL;
+  oparams->authid=NULL;
   oparams->realm=NULL;
   oparams->param_version=0;
 
@@ -1154,17 +1174,27 @@ static int c_continue_step (void *conn_context,
     /*nothing more to do; authenticated */
     oparams->doneflag=1;
     oparams->mech_ssf = 0;
-    oparams->authid = (char *) params->utils->malloc(strlen(text->authid) + 1);
-    strcpy(oparams->authid, text->authid);
 
-    oparams->user = (char *) params->utils->malloc(strlen(text->authid) + 1);
-    strcpy(oparams->user, text->authid);
+    if (cram_strdup(params->utils, text->authid, &(oparams->authid), NULL)!=SASL_OK)
+	return SASL_NOMEM;
+
+    if (cram_strdup(params->utils, text->authid, &(oparams->user), NULL)!=SASL_OK)
+	return SASL_NOMEM;
 
     VL(("clientout looks like=%s %i\n",*clientout,*clientoutlen));
 
     text->state++; /* fail if called again */
 
-    return SASL_OK;
+    return SASL_CONTINUE;
+  }
+
+  if (text->state == 3)
+  {
+      *clientout = NULL;
+      *clientoutlen = 0;
+      VL(("Verify we're done step"));
+      text->state++;
+      return SASL_OK;      
   }
 
   return SASL_FAIL; /* should never get here */
