@@ -1,28 +1,46 @@
 /* GSSAPI SASL plugin
  * Leif Johansson
  */
-/*****************************************************************
-        Copyright 1998 by Stockholm University 
-            and Carnegie Mellon University
 
-                      All Rights Reserved
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for any purpose and without fee is hereby granted,
-provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in
-supporting documentation, and that the name of CMU not be
-used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.
-
-CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
-ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
-CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
-ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
-ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
-SOFTWARE.
-******************************************************************/
+/* 
+ * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ */
 
 #include <config.h>
 
@@ -117,7 +135,7 @@ sasl_gss_disperr(context_t *context, char **outp,
 	      strcat(out, (char *) msg.value);
 	      strcat(out, "; ");
 	  }
-	  (void) gss_release_buffer(&min_stat, &msg);
+	  gss_release_buffer(&min_stat, &msg);
 	  
 	  if (!msg_ctx)
 	       break;
@@ -173,7 +191,7 @@ sasl_gss_encode(void *context, const char *input, unsigned inputlen,
   if (GSS_ERROR(maj_stat))
     {
       if (output_token->value)
-	  free(output_token->value);
+	  gss_release_buffer(&min_stat, output_token);
       return SASL_FAIL;
     }
 
@@ -182,16 +200,22 @@ sasl_gss_encode(void *context, const char *input, unsigned inputlen,
       int len;
 
       *output = text->malloc(output_token->length + 4);
+      if (*output == NULL) {
+	  gss_release_buffer(&min_stat, output_token);
+	  return SASL_NOMEM;
+      }
       len = htonl(output_token->length);
       memcpy(*output, &len, 4);
       memcpy(*output + 4, output_token->value, output_token->length);
-      free(output_token->value);
   }
 
   if (outputlen) {
       *outputlen = output_token->length + 4;
   }
 
+  if (output_token->value)
+      gss_release_buffer(&min_stat, output_token);
+
   return SASL_OK;
 }
 
@@ -241,7 +265,7 @@ sasl_gss_decode(void *context, const char *input, unsigned inputlen,
 	    text->size = ntohl(text->size);
 	    text->cursize = 0;
 
-	    if (text->size > 0xFFFF || text->size == 0) return SASL_FAIL;
+	    if (text->size > 0xFFFF || text->size <= 0) return SASL_FAIL;
 
 	    if (text->bufsize < text->size + 5) {
 		text->buffer = text->realloc(text->buffer, text->size + 5);
@@ -289,14 +313,18 @@ sasl_gss_decode(void *context, const char *input, unsigned inputlen,
     if (GSS_ERROR(maj_stat))
     {
 	if (output_token->value)
-	    free(output_token->value);
+	    gss_release_buffer(&min_stat, output_token);
 	return SASL_FAIL;
     }
 
-    if (output_token->value && output)
-	*output = output_token->value;
     if (outputlen)
 	*outputlen = output_token->length;
+    if (output_token->value) {
+	if (output)
+	    *output = output_token->value;
+	else
+	    gss_release_buffer(&min_stat, output_token);
+    }
 
     /* reset for the next packet */
     text->size = -1;
@@ -310,6 +338,10 @@ sasl_gss_decode(void *context, const char *input, unsigned inputlen,
 	if (extra != NULL) {
 	    /* merge the two packets together */
 	    *output = text->realloc(*output, *outputlen + extralen);
+	    if (*output == NULL) {
+		text->free(extra);
+		return SASL_NOMEM;
+	    }
 	    memcpy(*output + *outputlen, extra, extralen);
 	    *outputlen += extralen;
 	    text->free(extra);
@@ -377,20 +409,31 @@ sasl_gss_free_context_contents(context_t *text)
 {
   OM_uint32 maj_stat, min_stat;
   
-  if (text->gss_ctx != GSS_C_NO_CONTEXT)
+  if (text->gss_ctx != GSS_C_NO_CONTEXT) {
     maj_stat = gss_delete_sec_context (&min_stat,&text->gss_ctx,GSS_C_NO_BUFFER);
+    text->gss_ctx = GSS_C_NO_CONTEXT;
+  }
   
-  if (text->client_name != GSS_C_NO_NAME)
+  if (text->client_name != GSS_C_NO_NAME) {
     maj_stat = gss_release_name(&min_stat,&text->client_name);
+    text->client_name = GSS_C_NO_NAME;
+  }
 
-  if (text->server_name != GSS_C_NO_NAME)
+  if (text->server_name != GSS_C_NO_NAME) {
     maj_stat = gss_release_name(&min_stat,&text->server_name);
+    text->server_name = GSS_C_NO_NAME;
+  }
   
-  if ( text->server_creds != GSS_C_NO_CREDENTIAL)
+  if ( text->server_creds != GSS_C_NO_CREDENTIAL) {
     maj_stat = gss_release_cred(&min_stat, &text->server_creds);
+    text->server_creds = GSS_C_NO_CREDENTIAL;
+  }
 
   /* if we've allocated space for decryption, free it */
-  if (text->buffer) text->free(text->buffer);
+  if (text->buffer) {
+      text->free(text->buffer);
+      text->buffer = NULL;
+  }
    
 }
 
@@ -461,6 +504,7 @@ sasl_gss_server_step (void *conn_context,
 				      &text->server_name);
 	  
 	  params->utils->free(name_token.value);
+	  name_token.value = NULL;
 	  
 	  if (GSS_ERROR(maj_stat)) {
 	      sasl_gss_set_error(text, errstr, "gss_import_name",
@@ -468,7 +512,12 @@ sasl_gss_server_step (void *conn_context,
 	      sasl_gss_free_context_contents(text);
 	      return SASL_FAIL;
 	  }
-	  
+
+	  if ( text->server_creds != GSS_C_NO_CREDENTIAL) {
+	      maj_stat = gss_release_cred(&min_stat, &text->server_creds);	  
+	      text->server_creds = GSS_C_NO_CREDENTIAL;
+	  }
+
 	  maj_stat = gss_acquire_cred(&min_stat, 
 				      text->server_name,
 				      GSS_C_INDEFINITE, 
@@ -510,16 +559,23 @@ sasl_gss_server_step (void *conn_context,
       if (GSS_ERROR(maj_stat)) {
 	  sasl_gss_set_error(text, errstr, "gss_accept_sec_context",
 			     maj_stat, min_stat);
-	  if (output_token->value)
-	    params->utils->free(output_token->value);
+	  if (output_token->value) {
+	      gss_release_buffer(&min_stat, output_token);
+	  }
+
 	  sasl_gss_free_context_contents(text);
-	  return SASL_FAIL;
+	  return SASL_BADAUTH;
       }
       
-      if (serverout && output_token->length)
-	*serverout = output_token->value;
       if (serveroutlen)
-	*serveroutlen = output_token->length;
+	  *serveroutlen = output_token->length;
+      if (output_token->value) {
+	  if (serverout)
+	      *serverout = output_token->value;
+	  else
+	      gss_release_buffer(&min_stat, output_token);
+      }
+
       
       if (maj_stat == GSS_S_COMPLETE)
 	{
@@ -551,8 +607,11 @@ sasl_gss_server_step (void *conn_context,
 	if (GSS_ERROR(maj_stat)) {
 	    sasl_gss_set_error(text, errstr, "gss_display_name",
 			       maj_stat, min_stat);
+	    params->utils->free(name_without_realm.value);
 	    if (name_token.value)
-		params->utils->free(name_token.value);
+		gss_release_buffer(&min_stat, &name_token);
+	    if (without)
+		gss_release_name(&min_stat, &without);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_BADAUTH;
 	}
@@ -580,8 +639,11 @@ sasl_gss_server_step (void *conn_context,
 	    if (GSS_ERROR(maj_stat)) {
 		sasl_gss_set_error(text, errstr, "gss_display_name",
 				   maj_stat, min_stat);
+		params->utils->free(name_without_realm.value);
 		if (name_token.value)
-		    params->utils->free(name_token.value);
+		    gss_release_buffer(&min_stat, &name_token);
+		if (without)
+		    gss_release_name(&min_stat, &without);
 		sasl_gss_free_context_contents(text);
 		return SASL_BADAUTH;
 	    }
@@ -590,25 +652,40 @@ sasl_gss_server_step (void *conn_context,
 					text->client_name,
 					without,
 					&equal);
+    
 	    if (GSS_ERROR(maj_stat)) {
 		sasl_gss_set_error(text, errstr, "gss_display_name",
 				   maj_stat, min_stat);
+		params->utils->free(name_without_realm.value);
 		if (name_token.value)
-		    params->utils->free(name_token.value);
+		    gss_release_buffer(&min_stat, &name_token);
+		if (without)
+		    gss_release_name(&min_stat, &without);
 		sasl_gss_free_context_contents(text);
 		return SASL_BADAUTH;
 	    }
 
+	    gss_release_name(&min_stat,&without);
+
 	} else {
 	    equal = 0;
 	}
 
 	if (equal == 1) /* xxx True doesn't seem to exist in gssapi.h */
 	{
-	    oparams->authid = (char *)name_without_realm.value;
+	    oparams->authid = (char *) params->utils->malloc(strlen(name_without_realm.value)+1);
+	    if (oparams->authid == NULL) return SASL_NOMEM;
+	    strcpy(oparams->authid, name_without_realm.value);
 	} else {
-	    oparams->authid = (char *)name_token.value;	    
+	    oparams->authid = (char *) params->utils->malloc(strlen(name_token.value)+1);
+	    if (oparams->authid == NULL) return SASL_NOMEM;
+	    strcpy(oparams->authid, name_token.value);
 	}
+
+	if (name_token.value)
+	    params->utils->free(name_token.value);
+	if (name_without_realm.value)
+	    gss_release_buffer(&min_stat, &name_without_realm);
 	
 	/* we have to decide what sort of encryption/integrity/etc.,
 	   we support */
@@ -653,16 +730,21 @@ sasl_gss_server_step (void *conn_context,
 	    sasl_gss_set_error(text, errstr, "gss_wrap",
 			       maj_stat, min_stat);
 	    if (output_token->value)
-		params->utils->free(output_token->value);
+		gss_release_buffer(&min_stat, output_token);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
-	
-	if (serverout && output_token->length)
-	  *serverout = output_token->value;
+
+
 	if (serveroutlen)
-	  *serveroutlen = output_token->length;
-	
+            *serveroutlen = output_token->length;
+        if (output_token->value) {
+	    if (serverout)
+	        *serverout = output_token->value;
+	    else
+	        gss_release_buffer(&min_stat, output_token);
+        }
+
 	VL(("Sending %d bytes (ssfcap) to client\n",*serveroutlen));
         /* Wait for ssf request and authid */
 	text->state = SASL_GSSAPI_STATE_SSFREQ; 
@@ -688,8 +770,6 @@ sasl_gss_server_step (void *conn_context,
 	if (GSS_ERROR(maj_stat)) {
 	    sasl_gss_set_error(text, errstr, "gss_unwrap",
 			       maj_stat, min_stat);
-	    if (output_token->value)
-		params->utils->free(output_token->value);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
@@ -714,6 +794,8 @@ sasl_gss_server_step (void *conn_context,
 	    params->utils->log(params->utils->conn, SASL_LOG_WARNING,
 			       "GSSAPI", SASL_FAIL, 0, 
 		    "protocol violation: client requested invalid layer");
+	    if (output_token->value)
+		gss_release_buffer(&min_stat, output_token);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
@@ -739,14 +821,19 @@ sasl_gss_server_step (void *conn_context,
 	    oparams->user = user;
 	}
 
-	params->utils->free(output_token->value);
+	gss_release_buffer(&min_stat, output_token);
 	
 	text->state = SASL_GSSAPI_STATE_AUTHENTICATED;
-	
+
+	*serverout = NULL;
+	*serveroutlen = 0;	
+
 	return SASL_OK;
 	break;
       }
     case SASL_GSSAPI_STATE_AUTHENTICATED:
+	*serverout = NULL;
+	*serveroutlen = 0;
       return SASL_OK;
       break;
 
@@ -932,10 +1019,10 @@ get_userid(sasl_client_params_t *params,char **userid,sasl_interact_t **prompt_n
   if (prompt!=NULL)
     {
       /* copy it */
-      *userid=params->utils->malloc(strlen(prompt->result)+1);
+      *userid=params->utils->malloc(prompt->len+1);
       if ((*userid)==NULL) return SASL_NOMEM;
-	
-      strcpy(*userid, prompt->result);
+
+      strncpy(*userid, prompt->result, prompt->len+1);      
       return SASL_OK;
     }
 
@@ -952,8 +1039,11 @@ get_userid(sasl_client_params_t *params,char **userid,sasl_interact_t **prompt_n
 			NULL);
     if (result != SASL_OK)
       return result;
-    if (! id)
-      return SASL_BADPARAM;
+    if (! id) {
+	*userid = NULL;
+	return SASL_BADPARAM;
+    }
+
     *userid = params->utils->malloc(strlen(id) + 1);
     if (! *userid)
       return SASL_NOMEM;
@@ -980,12 +1070,12 @@ get_authid(sasl_client_params_t *params,
   prompt=find_prompt(prompt_need,SASL_CB_AUTHNAME);
   if (prompt!=NULL)
   {
-    /* copy it */
-    *authid=params->utils->malloc(strlen(prompt->result)+1);
-    if ((*authid)==NULL) return SASL_NOMEM;
+      /* copy it */
+      *authid=params->utils->malloc(prompt->len+1);
+      if ((*authid)==NULL) return SASL_NOMEM;
 
-    strcpy(*authid, prompt->result);
-    return SASL_OK;
+      strncpy(*authid, prompt->result, prompt->len+1);      
+      return SASL_OK;
   }
 
   /* Try to get the callback... */
@@ -1042,8 +1132,10 @@ sasl_gss_client_step (void *conn_context,
     {
     case SASL_GSSAPI_STATE_AUTHNEG:
       {
-	/* try to get the userid */
-	if (oparams->user==NULL)
+	  VL(("sasl_gss_client_step: AUTHNEG\n"));
+      
+	  /* try to get the userid */
+	  if (oparams->user==NULL)
 	  {
 	    int auth_result = SASL_OK;
 	    VL (("Trying to get userid\n"));
@@ -1051,7 +1143,9 @@ sasl_gss_client_step (void *conn_context,
 				   &oparams->user,
 				   prompt_need);
 	    
-	    VL (("Userid: %s\n",oparams->user));
+	    if (oparams->user) {
+		VL (("Userid: %s\n",oparams->user));
+	    }
 
 	    if ((auth_result!=SASL_OK) && (auth_result!=SASL_INTERACT))
 	      {
@@ -1094,6 +1188,7 @@ sasl_gss_client_step (void *conn_context,
 					&text->server_name);
 	    
 	    params->utils->free(name_token.value);
+	    name_token.value = NULL;
 	    
 	    if (GSS_ERROR(maj_stat))
 	      {
@@ -1116,7 +1211,7 @@ sasl_gss_client_step (void *conn_context,
            maj_stat = gss_delete_sec_context (&min_stat,&text->gss_ctx,GSS_C_NO_BUFFER);
            text->gss_ctx = GSS_C_NO_CONTEXT;
          }
-	
+
 	maj_stat =
 	  gss_init_sec_context(&min_stat,
 			       GSS_C_NO_CREDENTIAL,
@@ -1135,15 +1230,19 @@ sasl_gss_client_step (void *conn_context,
 	if (GSS_ERROR(maj_stat))
 	  {
 	    if (output_token->value)
-	      params->utils->free(output_token->value);
+		gss_release_buffer(&min_stat, output_token);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	  }
 	
-	if (clientout && output_token->length)
-	  *clientout = output_token->value;
 	if (clientoutlen)
-	  *clientoutlen = output_token->length;
+            *clientoutlen = output_token->length;
+        if (output_token->value) {
+	    if (clientout)
+	        *clientout = output_token->value;
+	    else
+	        gss_release_buffer(&min_stat, output_token);
+        }
 
 	if (maj_stat == GSS_S_COMPLETE)
 	  {
@@ -1161,6 +1260,8 @@ sasl_gss_client_step (void *conn_context,
 	sasl_ssf_t need, allowed;
 	char serverhas, mychoice;
 
+	VL(("sasl_gss_client_step: SSFCAP\n"));
+
 	real_input_token.value = (void *) serverin;
 	real_input_token.length = serverinlen;
 
@@ -1173,6 +1274,8 @@ sasl_gss_client_step (void *conn_context,
 	
 	if (GSS_ERROR(maj_stat)) {
 	    sasl_gss_free_context_contents(text);
+	    if (output_token->value)
+		gss_release_buffer(&min_stat, output_token);
 	    return SASL_FAIL;
 	}
 	
@@ -1216,10 +1319,13 @@ sasl_gss_client_step (void *conn_context,
 	    return SASL_TOOWEAK;
 	}
 	
-	params->utils->free(output_token->value);
-	output_token->value = NULL;
+	gss_release_buffer(&min_stat, output_token);
+
+	if (oparams->user)
+		alen = strlen(oparams->user);
+	else
+		alen = 0;
 
-	alen = strlen(oparams->user);
 	input_token->length = 4 + alen;
 	input_token->value = (char *)params->utils->malloc( (input_token->length + 1)* sizeof(char));
 	if (input_token->value == NULL)
@@ -1228,7 +1334,8 @@ sasl_gss_client_step (void *conn_context,
 	    return SASL_NOMEM;
 	  }
 	VL(("user: %s,buflen=%d\n",oparams->user,input_token->length));
-	memcpy((char *)input_token->value+4,oparams->user,alen);
+	if (oparams->user)
+	    memcpy((char *)input_token->value+4,oparams->user,alen);
 	
 	
 	((unsigned char *)input_token->value)[0] = mychoice;
@@ -1246,19 +1353,24 @@ sasl_gss_client_step (void *conn_context,
 			     output_token);
 	
 	params->utils->free(input_token->value);
+	input_token->value = NULL;
 	
 	if (GSS_ERROR(maj_stat))
 	  {
 	    if (output_token->value)
-	      params->utils->free(output_token->value);
+		gss_release_buffer(&min_stat, output_token);
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	  }
 	
-	if (clientout && output_token->length)
-	  *clientout = output_token->value;
 	if (clientoutlen)
-	  *clientoutlen = output_token->length;
+            *clientoutlen = output_token->length;
+        if (output_token->value) {
+	    if (clientout)
+	        *clientout = output_token->value;
+	    else
+	        gss_release_buffer(&min_stat, output_token);
+        }
 	
 	text->state = SASL_GSSAPI_STATE_AUTHENTICATED;
 	
@@ -1267,6 +1379,7 @@ sasl_gss_client_step (void *conn_context,
 	break;
       }
     case SASL_GSSAPI_STATE_AUTHENTICATED:
+      VL(("sasl_gss_client_step: AUTHENTICATED\n"));
       return SASL_OK;
       break;
 
@@ -1276,6 +1389,7 @@ sasl_gss_client_step (void *conn_context,
     }
 
   /* we should never get here!!! */
+  return SASL_FAIL;
 }
 
 static const sasl_client_plug_t client_plugins[] = 
